from typing import Any, Optional

from google.adk.agents.base_agent import BaseAgent
from google.adk.agents.callback_context import CallbackContext
from google.adk.plugins.base_plugin import BasePlugin
from google.adk.tools.base_tool import BaseTool
from google.adk.tools.tool_context import ToolContext
from google.genai import types

from .models import OPADKRequestInput, OPADKResponse
from .opa import OPABaseClient


class OPADKPlugin(BasePlugin):
    """
    Plugin that integrates OPA policy checks into agent and tool invocations.
    """

    opa_client: OPABaseClient

    def __init__(
        self,
        opa_client: OPABaseClient,
    ) -> None:
        super().__init__(name="opadk_plugin")
        self.opa_client = opa_client

    async def before_agent_callback(
        self, agent: BaseAgent, callback_context: CallbackContext
    ) -> Optional[types.Content]:
        outcome = await self.opa_client.is_allowed(
            scope="agent",
            input=OPADKRequestInput.from_callback_context(
                callback_context=callback_context,
            ).model_dump(),
        )

        if not outcome.allow:
            return self.agent_access_denied_response(outcome)

    async def before_tool_callback(
        self,
        *,
        tool_context: ToolContext,
        tool: BaseTool,
        tool_args: dict[str, Any],
    ) -> Optional[dict[str, Any]]:
        outcome = await self.opa_client.is_allowed(
            scope="tool",
            input=OPADKRequestInput.from_tool_callback(
                tool_context=tool_context,
                tool=tool,
                tool_args=tool_args,
            ).model_dump(),
        )

        if not outcome.allow:
            return self.tool_access_denied_response(outcome)

    def tool_access_denied_response(self, outcome: OPADKResponse) -> dict[str, Any]:
        """
        Response when tool access is denied by policy.
        """
        return {
            "error": "Tool invocation not allowed by policy.",
            "reasons": outcome.deny.reasons,
        }

    def agent_access_denied_response(self, outcome: OPADKResponse) -> types.Content:
        """
        Response when agent access is denied by policy.
        """
        parts = ["Agent invocation not allowed by policy."]

        if outcome.deny.reasons:
            parts.append("Reasons:")
            for reason in outcome.deny.reasons:
                parts.append(f"- {reason}")

        return types.Content(
            parts=[
                types.Part.from_text(
                    text="\n".join(parts),
                ),
            ]
        )



import asyncio
import json
import subprocess
from typing import Any, Literal

import aiohttp
from pydantic import BaseModel

from .models import OPADKResponse

Scope = Literal["agent", "tool"]


class OPABaseClient(BaseModel):
    async def is_allowed(self, scope: Scope, input: dict[str, Any]) -> OPADKResponse:
        raise NotImplementedError()


class OPARemoteClient(OPABaseClient):
    """
    OPA client that uses the eval API to query a remote OPA server.
    """

    server_url: str
    namespace: list[str] = ["adk"]

    async def is_allowed(self, scope: Scope, input: dict[str, Any]) -> OPADKResponse:
        async with aiohttp.ClientSession(base_url=self.server_url) as session:
            url = f"/v1/data/{'/'.join(self.namespace)}/{scope}"
            async with session.post(
                url,
                json={"input": input},
            ) as response:
                json = await response.json()
                return OPADKResponse.model_validate(json.get("result"))
        raise Exception("OPA query failed")


class OPARunClient(OPABaseClient):
    """
    OPA client that runs the `opa eval` command locally for development/testing.
    """

    opa_path: str = "opa"
    bundle_path: str
    namespace: list[str] = ["adk"]

    async def is_allowed(self, scope: Scope, input: dict[str, Any]) -> OPADKResponse:
        query = ".".join(["data", *self.namespace, scope])
        process = await asyncio.create_subprocess_exec(
            self.opa_path,
            "eval",
            "--format",
            "json",
            "--bundle",
            self.bundle_path,
            "--stdin-input",
            query,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )

        stdout, _stderr = await process.communicate(json.dumps(input).encode())

        try:
            data = json.loads(stdout.decode())
        except Exception as e:
            raise Exception(f"Failed to decode OPA response: {e}")

        if data.get("errors"):
            raise Exception(f"OPA returned errors: {data['errors']}")

        result = data.get("result")
        if result and len(result) > 0:
            result = result[0]
            expressions = result.get("expressions")
            if expressions and len(expressions) > 0:
                value = expressions[0].get("value")
                return OPADKResponse(**value)
        raise Exception("OPA eval failed")



from typing import Any, Optional

from google.adk.agents.callback_context import CallbackContext
from google.adk.events import Event
from google.adk.tools.base_tool import BaseTool
from google.adk.tools.tool_context import ToolContext
from pydantic import BaseModel


class OPADKDenyResponse(BaseModel):
    reasons: list[str] = []


class OPADKResponse(BaseModel):
    allow: bool = False
    deny: OPADKDenyResponse = OPADKDenyResponse()


class OPADKRequestAgent(BaseModel):
    name: str


class OPADKRequestTool(BaseModel):
    name: str
    args: dict[str, Any]


class OPADKRequestInput(BaseModel):
    """
    Input schema for OPA policy evaluation.
    """

    state: dict[str, Any]
    agent: OPADKRequestAgent
    tool: Optional[OPADKRequestTool] = None
    events: Optional[list[dict[str, Any]]] = None

    @classmethod
    def from_callback_context(
        cls,
        callback_context: CallbackContext,
    ) -> "OPADKRequestInput":
        return cls(
            state=callback_context.state.to_dict(),
            agent=OPADKRequestAgent(name=callback_context.agent_name),
            events=_serialize_events(callback_context.session.events),
        )

    @classmethod
    def from_tool_callback(
        cls,
        tool_context: ToolContext,
        tool: BaseTool,
        tool_args: dict[str, Any],
    ) -> "OPADKRequestInput":
        return cls(
            state=tool_context.state.to_dict(),
            agent=OPADKRequestAgent(name=tool_context.agent_name),
            tool=OPADKRequestTool(
                name=tool.name,
                args=tool_args,
            ),
            events=_serialize_events(tool_context.session.events),
        )


def _serialize_events(events: list[Event]) -> list[dict[str, Any]]:
    return [
        e.model_dump(
            mode="json", exclude_none=True, exclude_defaults=True, exclude_unset=True
        )
        for e in events
    ]
